/*
	OpenGL Frambuffer Object (FBO) abstraction. Includes fp16 framebuffers

	MIT License

	Copyright (c) 2009, Markus Broecker <mbroecker@gmail.com>

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
 */
#include "Framebuffer.h"
#include "OpenGL.h"

using namespace mbroecker;

Framebuffer::Framebuffer(unsigned int width, unsigned int height, unsigned int cbCount, Framebuffer::Format f) :
	mWidth(width), mHeight(height), mColorbufferCount(cbCount)
{	

	GLenum INTERNAL_FORMAT = GL_RGBA8;
	if (f==FLOAT)
		INTERNAL_FORMAT=GL_RGBA32F_ARB;
	
	glGenFramebuffersEXT(1, &mBuffer);
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, mBuffer );

	mColorbuffers = new unsigned int[mColorbufferCount];
	
	// create the color buffers
	glGenTextures( mColorbufferCount, mColorbuffers );
	for (unsigned int i = 0; i < mColorbufferCount; ++i)
	{
		glBindTexture( GL_TEXTURE_2D, mColorbuffers[i] );
		glTexImage2D( GL_TEXTURE_2D, 0, INTERNAL_FORMAT, mWidth, mHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0 );
		
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); 
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR); 
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP); 
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		//glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE );
		glGenerateMipmapEXT( GL_TEXTURE_2D );
		glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
		
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT + i, GL_TEXTURE_2D, mColorbuffers[i], 0);
		
	}
	
	// create the depth buffer
	glGenRenderbuffersEXT( 1, &mDepthbuffer );
	glBindRenderbufferEXT( GL_RENDERBUFFER_EXT, mDepthbuffer );
	glRenderbufferStorageEXT( GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, mWidth, mHeight );
	glFramebufferRenderbufferEXT( GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, mDepthbuffer );
	
	
	validate();
	disable();
	
}

Framebuffer::~Framebuffer()
{
	glDeleteTextures( mColorbufferCount, mColorbuffers );
	delete[] mColorbuffers;
	
	glDeleteRenderbuffersEXT(1, &mDepthbuffer);
	glDeleteFramebuffersEXT(1, &mBuffer);
}

void Framebuffer::bind() const
{
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, mBuffer );
	glViewport(0, 0, mWidth, mHeight );
	
	
	if (mColorbufferCount > 1)
	{
		GLenum buffers[mColorbufferCount];
		for (int i = 0; i < mColorbufferCount; ++i)
			buffers[i] = GL_COLOR_ATTACHMENT0_EXT + i;
		
		glDrawBuffers(mColorbufferCount, buffers);
	}
	/*
	else
		glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
	*/
	glError();
}

void Framebuffer::disable()
{
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );
	glDrawBuffer(GL_BACK);
}

void Framebuffer::validate() const throw(std::string)
{
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, mBuffer );
	GLenum err = glCheckFramebufferStatusEXT( GL_FRAMEBUFFER_EXT );
	if (err != GL_FRAMEBUFFER_COMPLETE_EXT)
	{
		switch (err)
		{
			case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
				throw(std::string("Incomplete framebuffer"));
				break;
			case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
				throw(std::string("Wrong/incomplete size"));
				break;
			case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
				throw(std::string("Incomplete drawing buffers"));
				break;
			case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
				throw(std::string("Incomplete formats"));
				break;
			case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
				throw(std::string("Missing attachments"));
				break;
			case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
				throw(std::string("Unsupported"));
				break;
			default:
				throw(std::string("Unknown error"));
		}
	}
}
