/*
	OpenGL Frambuffer Object (FBO) abstraction. Includes fp16 framebuffers

	MIT License

	Copyright (c) 2009, Markus Broecker <mbroecker@gmail.com>

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
 */

#ifndef FRAMEBUFFER_INCLUDED
#define FRAMEBUFFER_INCLUDED

#include <string>

namespace mbroecker
{

/**	An OpenGL Framebuffer Object. It's able to hold multiple colorbuffers and
	has an attached depthbuffer. It can be bound, and disabled. After use, you 
	should call glViewport again to set the required viewport dimensions.
	@author mbroecker
*/
class Framebuffer
{
public:
	/// Specified the format. Either 8Bit RGBA or float16
	enum Format
	{
		RGBA,
		FLOAT
	};
	
	/// \name Construction
	/// \{
	
	/**	@param width Framebuffer dimension
		@param height Framebuffer dimension
		@param colorbuffercount Number of color buffers
		@param f Format for all colorbuffers inside this FBO
	*/
	Framebuffer(unsigned int width, unsigned int height, unsigned int colorbuffercount = 1, Format f = RGBA);
	virtual ~Framebuffer();
	
	/// \}
	
	/// \name Interface
	/// \{
	
	/**	Binds and enables this FBO, also sets the correct viewport dimensions and enables writing to the
		correct buffers (if applicable).
	*/
	void bind() const;
	
	/// Return to normal OpenGL backbuffer rendering
	static void disable();
	
	inline unsigned int getId() const { return mBuffer; }
	inline unsigned int getColorbufferCount() const { return mColorbufferCount; }
	
	/// No boundary checking is done!
	inline unsigned int getColorbuffer(unsigned int i) const { return mColorbuffers[i]; }
	inline unsigned int getWidth() const { return mWidth; }
	inline unsigned int getHeight() const { return mHeight; }
	
	/// validates the Framebuffer. Gets called automatically by the constructor
	void validate() const throw(std::string);
	
	/// \}
	
protected:
	unsigned int	mWidth, mHeight, mColorbufferCount;
	unsigned int*	mColorbuffers;
	unsigned int	mDepthbuffer;
	
	unsigned int 	mBuffer;

private:
	/// An FBO is non copyable
	Framebuffer(const Framebuffer& cp);
	
};

};

#endif
