/*
	GLSL Shader abstraction

	MIT License

	Copyright (c) 2009, Markus Broecker <mbroecker@gmail.com>

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
 */

#include "Shader.h"

#include <vector>
#include <fstream>
#include <iostream>
#include <cstring>

using namespace mbroecker;

std::string Shader::readFile(const std::string& filename) const
{
	std::ifstream file;
	file.open( filename.c_str() );
	
	if (!file.is_open())
	{
		std::cerr << "Error opening file " << filename << "!\n";
		return "";
	}
	
	std::string contents = "";
	std::string line;
	while (!file.eof())
	{
		std::getline( file, line );
		contents = contents + "\n" + line;	
	}
	
	file.close();
	
	return contents;
}

GLShader::GLShader( const std::string& vpFile, const std::string& fpFile)
{

	load( vpFile, fpFile );
}

GLShader::~GLShader()
{
	glDeleteProgram( mProgram );
	glDeleteShader( mVS );
	glDeleteShader( mFS );	
}

void GLShader::bind() const
{
	glUseProgram( mProgram );
}

void GLShader::disable() const
{
	glUseProgram( 0 );
}

int GLShader::getUniform(const std::string& name) const
{
	return glGetUniformLocation( mProgram, name.c_str() );
}

void GLShader::setUniform(const std::string& name, int i) const
{
	glUniform1i( glGetUniformLocation( mProgram, name.c_str() ), i );
}

void GLShader::setUniform(const std::string& name, float f) const
{
	glUniform1f( glGetUniformLocation( mProgram, name.c_str() ), f );
}

void GLShader::setUniform(const std::string& name, float x, float y, float z) const
{
	glUniform3f( glGetUniformLocation( mProgram, name.c_str() ), x, y, z );
}

void GLShader::setMatrix4(const std::string& name, const float* matrix) const
{
	glUniformMatrix4fv( glGetUniformLocation(mProgram, name.c_str()), 16, GL_FALSE, matrix);
}

void GLShader::setTexture2D(const std::string& name, unsigned int textureId, unsigned int textureUnit) const
{
	glActiveTexture(GL_TEXTURE0 + textureUnit);
	glBindTexture(GL_TEXTURE_2D, textureId);
	glUniform1i( glGetUniformLocation( mProgram, name.c_str() ), textureUnit );
}

void GLShader::setTextureCube(const std::string& name, unsigned int texId, unsigned int unit) const
{
	glActiveTexture(GL_TEXTURE0 + unit);
	glBindTexture(GL_TEXTURE_CUBE_MAP, texId);
	glUniform1i( glGetUniformLocation( mProgram, name.c_str() ), unit );
}

void GLShader::loadShader(const std::string& fileName, GLenum type)
{
	std::vector<std::string>	contents;
	std::ifstream file;
	file.open( fileName.c_str() );
	
	if (!file.is_open())
	{
		std::cerr << "Error opening file " << fileName << "!\n";
		return;
	}
	
	std::string line;
	while (!file.eof())
	{
		std::getline( file, line );
		contents.push_back( line );	
	}
		
	file.close();

	int current = 0;
	if (type == GL_VERTEX_SHADER)
		current = mVS = glCreateShader( GL_VERTEX_SHADER );
	else
		current = mFS = glCreateShader( GL_FRAGMENT_SHADER );
		
	
	const char *glLines[256];
	for (unsigned int i = 0; i < contents.size(); ++i)
		glLines[i] = contents[i].c_str();
						
	glShaderSource( current, contents.size(), glLines , 0 );
	glCompileShader( current );
	
	const unsigned int BUFFER_SIZE = 512;
	char buffer[BUFFER_SIZE];
	memset( buffer, 0, BUFFER_SIZE );
	int length = 0;
	
	glGetShaderInfoLog( current, BUFFER_SIZE, &length, buffer );
	if (length > 0)
	{
		std::cerr << (type==GL_VERTEX_SHADER?"Vertex":"Fragment") << "shader compile error: " << buffer;
		std::cerr << std::endl;
	}
}

void GLShader::load(const std::string& vpFile, const std::string& fpFile)
{
	loadShader( vpFile, GL_VERTEX_SHADER );
	loadShader( fpFile, GL_FRAGMENT_SHADER );
	
	mProgram = glCreateProgram();
	glAttachShader( mProgram, mVS );
	glAttachShader( mProgram, mFS );;
	glLinkProgram( mProgram );

	
	const unsigned int BUFFER_SIZE = 512;
	char buffer[BUFFER_SIZE];
	memset( buffer, 0, BUFFER_SIZE );
	int length = 0;
	
	memset( buffer, 0, BUFFER_SIZE );
	glGetProgramInfoLog( mProgram, BUFFER_SIZE, &length, buffer );
	if (length > 0)
	{
		std::cerr << "Shader link error: " << buffer << std::endl;
	}
	
	glValidateProgram( mProgram );
	int status;
	glGetProgramiv( mProgram, GL_VALIDATE_STATUS, &status );
	if (status == GL_FALSE)
	{
		std::cerr << "Error validating shader\n";
	}
}

#ifdef CG_VERSION_NUM
#include <Cg/cgGL.h>

CGcontext CGShader::sContext = 0;
CGShader::CGShader(const std::string file, Profile p)
{
	if (!sContext)
	{
		std::clog << "Creating Cg Context ... \n";
		sContext = cgCreateContext();
	}
	
	if (p == VERTEX_SHADER)
		mProfile = CG_PROFILE_GLSLV;
	else
		mProfile = CG_PROFILE_GLSLF;
	
	std::string contents = readFile(file);
	
	//std::clog << contents << std::endl;
	
	mProgram = cgCreateProgram( sContext, CG_SOURCE, contents.c_str(), mProfile, "main", 0);
	checkError();
	
	cgGLLoadProgram( mProgram );
	checkError();
}

CGShader::~CGShader()
{
	cgDestroyProgram( mProgram );
}

void CGShader::bind() const
{
	cgGLEnableProfile( mProfile );
	cgGLBindProgram( mProgram );
	
	checkError();
}

void CGShader::disable() const
{
	cgGLDisableProfile( mProfile );
}

void CGShader::setUniform(const std::string& name, int i) const
{
	cgGLSetParameter1d( cgGetNamedParameter(mProgram, name.c_str()), i );
}

void CGShader::setUniform(const std::string& name, float f) const
{
	cgGLSetParameter1f( cgGetNamedParameter( mProgram, name.c_str()), f);
}

void CGShader::setUniform(const std::string& name, float x, float y, float z) const
{
	cgGLSetParameter3f( cgGetNamedParameter( mProgram, name.c_str()), x, y, z );
}

void CGShader::setTexture2D(unsigned int textureId, const std::string& textureName, unsigned int textureUnit) const
{
	cgGLSetupSampler( cgGetNamedParameter( mProgram, textureName.c_str() ), textureId);
}

void CGShader::checkError() const
{
	CGerror error = cgGetError();
	if (error != CG_NO_ERROR)
	{
		const char* errorString = cgGetErrorString(error);
		std::cerr << "CG error: " << errorString << std::endl;
	}
}

#endif
