/*
	GLSL Shader abstraction
 
	MIT License

	Copyright (c) 2009, Markus Broecker <mbroecker@gmail.com>

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
 */

#ifndef SHADER_INCLUDED
#define SHADER_INCLUDED

#include <string>
#include "OpenGL.h"


namespace mbroecker
{
	
class Shader
{
public:
	/// \name Destruction
	/// \{
	Shader() {};
	virtual ~Shader() {};
	/// \}
	
	/// \name Interface
	/// \{
	virtual void bind() const = 0;
	virtual void disable() const = 0;
	
	
	virtual void setUniform(const std::string& name, int i) const = 0;
	virtual void setUniform(const std::string& name, float f) const = 0;
	virtual void setUniform(const std::string& name, float x, float y, float z) const = 0;
	virtual void setMatrix4(const std::string& name, const float* matrix) const = 0;
	virtual void setTexture2D(const std::string& textureName, unsigned int textureId,unsigned int textureUnit=0) const = 0;
	virtual void setTextureCube(const std::string& textureName, unsigned int textureId, unsigned int textureUnit=0) const = 0;
	
	/// \}

protected:
	std::string readFile(const std::string& filename) const;
	
private:
	Shader(const Shader& sh);
	
};

class GLShader : public Shader
{
public:
	GLShader(const std::string& vpFile, const std::string& fpFile);
	~GLShader();
	
	virtual void bind() const;
	virtual void disable() const;
	virtual int getUniform(const std::string& name) const;	
	
	virtual void setUniform(const std::string& name, int i) const;
	virtual void setUniform(const std::string& name, float f) const;
	virtual void setUniform(const std::string& name, float x, float y, float z) const;
	virtual void setMatrix4(const std::string& name, const float* matrix) const;
	virtual void setTexture2D(const std::string& textureName, unsigned int textureId, unsigned int textureUnit=0) const;
	virtual void setTextureCube(const std::string& textureName, unsigned int textureId, unsigned int textureUnit=0) const;

private:
	int				mVS, mFS, mProgram;		
	
	void load(const std::string& vertexFile, const std::string& fragmentFile);
	void loadShader(const std::string& fileName, GLenum type);
};

#ifdef CG_VERSION_NUM
#include "OpenGL.h"
#include <Cg/cg.h>

class CGShader : public Shader
{
public:
	enum Profile
	{
		VERTEX_SHADER,
		FRAGMENT_SHADER
	};
	
	CGShader(const std::string file, Profile p);
	~CGShader();
	
	void bind() const;
	void disable() const;
	
	virtual void setUniform(const std::string& name, int i) const;
	virtual void setUniform(const std::string& name, float f) const;
	virtual void setUniform(const std::string& name, float x, float y, float z) const ;
	virtual void setTexture2D(unsigned int textureId, const std::string& textureName, unsigned int textureUnit=0) const;
	
	
private:
	static CGcontext	sContext;
	CGprogram			mProgram;
	CGprofile			mProfile;
	
	void checkError() const;
};

#endif
	
};

#endif
