/*
	Deferred shading example

	MIT License

	Copyright (c) 2009, Markus Broecker <mbroecker@gmail.com>

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/

#include <cstdlib>
#include <cmath>
#include <iostream>
#include "OpenGL.h"
#include "Framebuffer.h"
#include "Shader.h"

using namespace mbroecker;

static float camera_anim = 0.f;
static unsigned int oldTime = 0;
static int gMainWindow = 0;

static Framebuffer*	gRenderTarget;
static Framebuffer*	gDisplayTarget;

static Shader*		gTonemapper;

static Shader*		gDeferred;
static Shader*		gAssembler;

enum DisplayMode
{
	ASSEMBLED,
	POSITION,
	NORMAL,
	ALBEDO
} gDisplayMode = ASSEMBLED;


// renders a single quad with texture coordinates
static void renderQuad()
{
	glBegin(GL_QUADS);
	glTexCoord2i(0, 1);
	glVertex2i(-1, 1);
	glTexCoord2i(0, 0);
	glVertex2i(-1, -1);
	glTexCoord2i(1, 0);
	glVertex2i(1, -1);
	glTexCoord2i(1, 1);
	glVertex2i(1, 1);
	glEnd();
}

// renders the scene
static void renderScene()
{
	/*	The current, unaltered modelview matrix contains the view matrix. We have to 
		split those two matrices for deferred shading, as we will do lighting in world
		coordinates (where it belongs) instead of eye coordinates.
	*/
	static float viewMatrix[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, viewMatrix);
	gDeferred->setMatrix4("viewMatrix", viewMatrix);
	glError();
	
	/*	Create the world matrix for the teapot. We use the OpenGL matrix multiplications
		for convenience.
	*/
	glPushMatrix();
	glLoadIdentity();
	
	// GLUT's teapot is 'a little off' ... so compensate for that
	glTranslatef( 0.f, -0.7f, 0.f );
	glRotatef( -90, 1, 0, 0 );
	glScalef( 0.5, 0.5, 0.5 );
	
	//	We use this array to store model matrices
	static float modelMatrix[16];
	
	glGetFloatv(GL_MODELVIEW_MATRIX, modelMatrix);
	gDeferred->setMatrix4("modelMatrix", modelMatrix);
	glPopMatrix();
	
	gDeferred->setUniform("color", 0.6f, 0.6f, 0.1f);
	gDeferred->setUniform("gloss", 0.5f );
	glutSolidTeapot( 1.0 );
	glError();
	
	// create an array of boxes
	for (int x = -20; x <= 20; x += 4)
	{
		for (int z = -20; z <= 20; z += 4)
		{
			if (x == 0 && z == 0)
				continue;
			
			// again, construct the modelmatrices
			glPushMatrix();
			glLoadIdentity();
			glTranslatef( x, 0.f, z );
			
			glGetFloatv(GL_MODELVIEW_MATRIX, modelMatrix);
			gDeferred->setMatrix4("modelMatrix", modelMatrix);
			glPopMatrix();
					
			gDeferred->setUniform("color", 0.6f, 0.1f, 0.1f);
			gDeferred->setUniform("gloss", 0.2f);
			glutSolidCube( 1.5f );
		}
	}
	
	// and render a ground plane
	glPushMatrix();
	glLoadIdentity();
	glTranslatef( 0, -1, 0 );
	glGetFloatv(GL_MODELVIEW_MATRIX, modelMatrix);
	gDeferred->setMatrix4("modelMatrix", modelMatrix);
	glPopMatrix();
	
	gDeferred->setUniform("color", 0.2f, 0.2f, 0.2f);
	gDeferred->setUniform("gloss", 0.f);
	glBegin(GL_QUADS);
	glNormal3f( 0.f, 1.f, 0.f );
	glVertex3i( -40, 0, -40 );
	glVertex3i( -40, 0, 40 );
	glVertex3f( 40, 0, 40 );
	glVertex3f( 40, 0, -40 );
	glEnd();
	
}

static void display()
{	
	// setup projection
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, 1.0, 1.0, 100.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// let the camera rotate around the center
	float x = sin(camera_anim) * 6.f;
	float z = cos(camera_anim) * 6.f;
	gluLookAt(x, 2.0, z, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
	
	// setup scene rendering with the deferred shader
	gDeferred->bind();
	gRenderTarget->bind();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glEnable(GL_DEPTH_TEST);
	renderScene();
	
	gRenderTarget->disable();
	gDeferred->disable();

	
	// setup orthographic projection for fullscreen quad rendering
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.0, 1.0, -1.0, 1.0, 0.0, 1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glError();
	
	// instead of clearing the depth buffer every time, just disable depth testing ...
	glDisable(GL_DEPTH_TEST);
	
	
	/*	Assemble the deferred shading information, that is: shade the
		final picture in the fragment shader.
	*/
	gDisplayTarget->bind();
	gAssembler->bind();
	
	gAssembler->setTexture2D("position", gRenderTarget->getColorbuffer(0), 0);
	gAssembler->setTexture2D("normal", gRenderTarget->getColorbuffer(1), 1);
	gAssembler->setTexture2D("albedo", gRenderTarget->getColorbuffer(2), 2);
	renderQuad();
	glError();
	
	gAssembler->disable();
	gDisplayTarget->disable();
	
	/*	Render the final outcome. Our 'tonemapper' is a simple pass-through shader
		in this example.
	*/
	glViewport(0, 0, 512, 512);
	gTonemapper->bind();
	
	// set the texture depending on the user's choice
	unsigned int texture = 0;
	switch (gDisplayMode)
	{
		case ASSEMBLED:
			texture = gDisplayTarget->getColorbuffer(0);
			break;
		case POSITION:
			texture = gRenderTarget->getColorbuffer(0);
			break;
		case NORMAL:
			texture = gRenderTarget->getColorbuffer(1);
			break;
		case ALBEDO:
			texture = gRenderTarget->getColorbuffer(2);
			break;
	}
	
	gTonemapper->setTexture2D("texture", texture, 0);
	renderQuad();
	glError();
	
	gTonemapper->disable();
	
	// end this frame
	glutSwapBuffers();
}

static void idle()
{
	// make sure the main window is active
	if (glutGetWindow() != gMainWindow)
		glutSetWindow(gMainWindow);

	unsigned int time = glutGet(GLUT_ELAPSED_TIME);
	float delta = (float)(time-oldTime) / 1000.f;
	oldTime = time;
	
	camera_anim += delta;
	
	glutPostRedisplay();
}

static void keyboard(unsigned char c, int x, int y)
{
	switch (c)
	{
		case 27:
			exit(0);
			break;
			
		// let the user decide which texture he wants to see
		case '1':
			std::cout << "Displaymode: Assembled\n";
			gDisplayMode = ASSEMBLED;
			break;
		case '2':
			std::cout << "Displaymode: Position FB\n";
			gDisplayMode = POSITION;
			break;
		case '3':
			std::cout << "Displaymode: Normal FB\n";
			gDisplayMode = NORMAL;
			break;
		case '4':
			std::cout << "Displaymode: Albedo FB\n";
			gDisplayMode = ALBEDO;
			break;
			
			
		default:
			break;
	}
}

int main(int argc, char** argv)
{
	glutInitWindowSize(512, 512);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutInit(&argc, argv);
	
	gMainWindow = glutCreateWindow("Deferred Shading Example");
#ifndef __APPLE__
	glewInit(&argc, argv);
#endif
	
	glutDisplayFunc(display);
	glutIdleFunc(idle);
	glutKeyboardFunc(keyboard);
	
	glClearColor(0.1, 0.3, 0.6, 0.0);

	
	// setup shaders/framebuffers
	gRenderTarget = new Framebuffer(512, 512, 3, Framebuffer::FLOAT);
	gDeferred = new GLShader("deferred.vert", "deferred.frag");
	gAssembler = new GLShader("assembler.vert", "assembler.frag");
	
	gDisplayTarget = new Framebuffer(512, 512, 1, Framebuffer::FLOAT);
	gTonemapper = new GLShader("tonemapper.vert", "tonemapper.frag");
	
	oldTime = glutGet(GLUT_ELAPSED_TIME);
	glutMainLoop();
	
	return 0;
}
